#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <pthread.h>

#include "io_vanilla.h"
#include "generate_vanilla.h"
#include "stopwatch.h"

typedef double double3[3];

/*const double gamma = 6.674E-11;*/
const double G = 1.0;
const double epsilon2 = 1E-14;

/*for tiled version*/
const int B=50;
#define MIN(x,y) (((x)<(y)) ? (x) : (y))
#define MAX(x,y) (((x)>(y)) ? (x) : (y))

void acceleration_tiled (int n, double x[][3], double m[], double a[][3])
{
  int I,J,i,j,iend,jstart,jend;
  double d0,d1,d2,r,r2,factori,factorj,invfact;
  double X[B][6], A[B][6], M[B][2];

  /* compute acceleration exploiting symmetry */
  for (I=0; I<n; I+=B)
    for (J=I; J<n; J+=B) {
	  /* load to avoid aliasing */
	  iend = MIN(I+B,n);
	  jend = MIN(J+B,n);
	  for (i=I; i<iend; i++) {
	    X[i-I][0] = x[i][0]; X[i-I][1] = x[i][1]; X[i-I][2] = x[i][2];
	    A[i-I][0] = 0.0;     A[i-I][1] = 0.0;     A[i-I][2] = 0.0;
	    M[i-I][0] = m[i];
	  }
	  for (j=J; j<jend; j++) {
	    X[j-J][4] = x[j][0]; X[j-J][5] = x[j][1]; X[j-J][6] = x[j][2];
	    A[j-J][4] = 0.0;     A[j-J][5] = 0.0;     A[j-J][6] = 0.0;
	    M[j-J][1] = m[j];
	  }
	  /* compute locally */
	  for (i=0; i<iend-I; i++) {
	    jstart = MAX(J,I+i+1)-J; 
	    for (j=jstart; j<jend-J; j++) {
		  d0 = X[j][4]-X[i][0]; d1 = X[j][5]-X[i][1]; d2 = X[j][6]-X[i][2];
		  r2 = d0*d0 + d1*d1 + d2*d2 + epsilon2;
		  r = sqrt(r2);
		  invfact = G/(r*r2);
		  factori = M[i][0]*invfact;
		  factorj = M[j][1]*invfact;
		  A[i][0] += factorj*d0; A[i][1] += factorj*d1; A[i][2] += factorj*d2;
		  A[j][4] -= factori*d0; A[j][5] -= factori*d1; A[j][6] -= factori*d2;
		}
	  }
	  /* store result */
	  for (i=I; i<iend; i++) {
	    a[i][0] += A[i-I][0]; a[i][1] += A[i-I][1]; a[i][2] += A[i-I][2];
	  }
	  for (j=J; j<jend; j++) {
	    a[j][0] += A[j-J][4]; a[j][1] += A[j-J][5]; a[j][2] += A[j-J][6];
	  }
	}
}

void leapfrog (int n, double dt, double x[][3], double v[][3], double m[], 
			   double a[][3], double aold[][3])
{
  int i;
  double dt2 = dt*dt*0.5;
  double dthalf = dt*0.5;

  /* update position */
  for (i=0; i<n; i++)
    {
      x[i][0] += dt*v[i][0] + dt2*a[i][0];
      x[i][1] += dt*v[i][1] + dt2*a[i][1];
      x[i][2] += dt*v[i][2] + dt2*a[i][2];
    }

  /* save and clear acceleration */
  for (i=0; i<n; i++)
    {
      aold[i][0] = a[i][0];
      aold[i][1] = a[i][1];
      aold[i][2] = a[i][2];
      a[i][0] = a[i][1] = a[i][2] = 0.0;
    }
  
  /* compute new acceleration */
  acceleration_tiled(n,x,m,a);

  /* update velocity */
  for (i=0; i<n; i++)
    {
      v[i][0] += dthalf*aold[i][0] + dthalf*a[i][0];
      v[i][1] += dthalf*aold[i][1] + dthalf*a[i][1];
      v[i][2] += dthalf*aold[i][2] + dthalf*a[i][2];
    }
}
	       

int main (int argc, char** argv)
{
  int n;
  double3 *x;
  double3 *v;
  double *m;
  double3 *a;
  double3 *temp;
  int i,k,mod;
  char base[]="test1";
  char name[256];
  FILE *file;
  double t,dt;
  double start,stop;
  double elapsed,flop;
  int timesteps;

  if (argc!=4)
	{
	  printf("usage: vanilla <nbodies> <timesteps> <every>\n");
	  return 1;
	}
  sscanf(argv[1],"%d",&n);
  sscanf(argv[2],"%d",&timesteps);
  sscanf(argv[3],"%d",&mod);

  /* allocate memory, read data file */
  /*   file = fopen("plummer10000.vtk","r"); */
  /*   if (file==NULL) */
  /*     { */
  /*       printf("could not open file --- aborting\n"); */
  /*       return 1; */
  /*     } */
  /*   n = get_vtk_numbodies(file); */
  /*   rewind(file); */
  /*   x = calloc(n,sizeof(double3)); */
  /*   v = calloc(n,sizeof(double3)); */
  /*   m = calloc(n,sizeof(double)); */
  /*   read_vtk_file_double(file,n,x,v,m,&t,&dt); */
  /*   fclose(file); */
  /*   printf("loaded %d bodies\n",n); */
  x = calloc(n,sizeof(double3));
  v = calloc(n,sizeof(double3));
  m = calloc(n,sizeof(double));
  plummer(n,17,x,v,m);

  a = calloc(n,sizeof(double3));
  temp = calloc(n,sizeof(double3));
  for (i=0; i<n; i++)
    a[i][0] = a[i][1] = a[i][2] = 0.0;
  acceleration_tiled(n,x,m,a);

  k = 0;
  t = 0.0;
  dt = 1E-3;
  printf("writing %s_%06d.vtk \n",base,k);
  sprintf(name,"%s_%06d.vtk",base,k);
  file = fopen(name,"w");
  write_vtk_file_double(file,n,x,v,m,t,dt);
  fclose(file);
  start = get_time();

  for (k=1; k<timesteps; k++)
    {
      leapfrog(n,dt,x,v,m,a,temp);
      t += dt;
      if (k%mod==0)
		{
		  stop = get_time();
		  elapsed = stop-start;
		  flop = mod*(13.0*n*(n-1.0)+24.0*n);
		  printf("%g seconds for %g ops = %g MFLOPS\n",elapsed,flop,flop/elapsed/1E6);
		  printf("writing %s_%06d.vtk \n",base,k/mod);

		  sprintf(name,"%s_%06d.vtk",base,k/mod);
		  file = fopen(name,"w");
		  write_vtk_file_double(file,n,x,v,m,t,dt);
		  fclose(file);

		  start = get_time();
		}
    }

  return 0;
}
